// imports.gi
const { BindConstraint, BindCoordinate }  = imports.gi.Clutter
const { BindingFlags }                    = imports.gi.GObject
const { BlurMode }                        = imports.gi.Shell
const { Bin }                             = imports.gi.St

// local modules
import RepaintSignal                      from '../effect/repaint-signal.js'
import { Connections }                    from '../utils/connections.js'
import settings                           from '../utils/settings.js'
import * as UI                            from '../utils/ui.js'
import constants                          from '../utils/constants.js'
import { Padding }                        from '../utils/types.js'
import { RoundedCornersManager }          from './rounded-corners-manager.js'


// --------------------------------------------------------------- [end imports]

// Patched blur effect
let blurEffect = null

export class BlurEffectManager {
    constructor () {
        this.blur_actors = new Map ()
        this._enabled = false
        this.connections = new Connections ()
    }

    /** Called when extension is enable  */
    enable () {
        if (!this._check_lib ()) {
            return
        }

        this._enabled = true

        // try to add blur effect to windows have opened
        global.get_window_actors ().forEach ((a) => {
            this._should_enable_effect (a.meta_window) && this._add_effect (a)
        })

        const repaint = () => {
            this.blur_actors.forEach ((blur_actor, win) => {
                if (blur_actor.visible) {
                    blur_actor.get_effects ()[0].queue_repaint ()
                    win.get_compositor_private ().opacity =
                        settings ().blurred_window_opacity
                }
            })
        }

        const background = global.window_group.first_child
        const repaint_effect = new RepaintSignal ()
        background.add_effect_with_name ('Repaint Effect', repaint_effect)
        const c = this.connections

        c.connect (repaint_effect, 'repaint', () => repaint ())

        const wm = global.window_manager
        /** When window is opened  */
        c.connect (wm, 'map', (_, actor) => {
            if (this._should_enable_effect (actor.meta_window)) {
                this._add_effect (actor)
            }
        })

        /** When window is closed  */
        c.connect (wm, 'destroy', (_, actor) => {
            this._remove_effect (actor)
        })

        /** When window raised & down */
        c.connect (global.display, 'restacked', () => {
            this.blur_actors.forEach ((blur, win) => {
                if (!blur.visible) {
                    return
                }

                const actor = win.get_compositor_private ()
                global.window_group.set_child_below_sibling (blur, actor)
            })
        })

        /** When Settings has changed  */
        const gs = settings ().g_settings
        c.connect (gs, 'changed', (_, k) => {
            this._settings_changed (k)
        })
    }

    /** Called when extension is disable  */
    disable () {
        global.get_window_actors ().forEach ((a) => this._remove_effect (a))
        this.connections.disconnect_all ()
        global.window_group.first_child.remove_effect_by_name ('Repaint Effect')
        this._enabled = false
    }

    query_blur (win) {
        return this.blur_actors.get (win)
    }

    // ------------------------------------------------------- [private methods]

    _check_lib () {
        if (this._enabled) {
            return false
        }

        try {
            blurEffect = imports.gi.Patched.BlurEffect
        } catch (e) {
            return false
        }

        if (!blurEffect) {
            return false
        }
        return true
    }

    /** Decide wether should add blur effect to a window */
    _should_enable_effect (win) {
        if (!win) {
            return false
        }
        // Skip when window shouldn't have rounded corners effect
        if (!RoundedCornersManager.should_enable_effect (win)) {
            return false
        }
        const class_instance = win.get_wm_class_instance ()
        if (!class_instance) {
            return false
        }
        return settings ().blur_list.includes (class_instance)
    }

    /** Add Blur effect to window  */
    _add_effect (actor) {
        UI.WhenSurfaceActorIsReady (this.connections, actor, () => {
            const blur_actor = new Bin ()
            blur_actor.set_name ('Blur Actor')

            // Bind props

            const flag = BindingFlags.SYNC_CREATE
            for (const prop of [
                'pivot-point',
                'visible',
                'translation-x',
                'translation-y',
                'scale-x',
                'scale-y',
            ]) {
                actor.bind_property (prop, blur_actor, prop, flag)
            }

            // Bind Position and Size of blur actor
            const coordinates = [
                BindCoordinate.X,
                BindCoordinate.Y,
                BindCoordinate.WIDTH,
                BindCoordinate.HEIGHT,
            ]
            for (const coordinate of coordinates) {
                blur_actor.add_constraint (
                    new BindConstraint ({
                        source: actor,
                        coordinate,
                        offset: 0,
                    })
                )
            }

            this.blur_actors.set (actor.meta_window, blur_actor)

            if (!blurEffect) {
                return
            }
            const effect = new blurEffect ()
            effect.mode = BlurMode.BACKGROUND
            blur_actor.add_effect_with_name (constants.BLUR_EFFECT, effect)

            global.window_group.insert_child_below (blur_actor, actor)

            this.update_coordinates (actor.meta_window)
            this.update_blur_effect (actor)

            this.connections.connect (actor.meta_window, 'size-changed', () => {
                this.update_coordinates (actor.meta_window)
                actor.opacity = settings ().blurred_window_opacity
            })
            this.connections.connect (actor, 'damaged', () => {
                actor.opacity = settings ().blurred_window_opacity
                this.connections.disconnect (actor, 'damage')
            })
        })
    }

    _remove_effect (actor) {
        this.connections.disconnect_all (actor)
        this.connections.disconnect_all (actor.meta_window)

        const blur_actor = this.blur_actors.get (actor.meta_window)
        if (blur_actor) {
            actor.opacity = 255

            blur_actor.destroy ()
            this.blur_actors.delete (actor.meta_window)
        }
    }

    compute_offset_of_blur (win) {
        const blur_actor = this.blur_actors.get (win)
        if (!blur_actor) {
            return [0, 0, 0, 0]
        }

        const cfg = UI.ChoiceRoundedCornersCfg (
            settings ().global_rounded_corner_settings,
            settings ().custom_rounded_corner_settings,
            win
        )
        const should_have_padding =
            cfg.keep_rounded_corners ||
            !(
                win.maximized_horizontally ||
                win.maximized_vertically ||
                win.fullscreen
            )

        const [x, y, width, height] = UI.computeWindowContentsOffset (win)

        // Set padding to zero when window should skip rounded corners
        const padding = should_have_padding ? cfg.padding : new Padding ()

        const scale = UI.scaleFactor ()

        return [
            x + padding.left * scale,
            y + padding.top * scale,
            width - (padding.left + padding.right + 1) * scale,
            height - (padding.top + padding.bottom + 1) * scale,
        ]
    }

    update_coordinates (win) {
        const blur_actor = this.blur_actors.get (win)
        if (!blur_actor) {
            return
        }

        const constraints = blur_actor.get_constraints ()
        this.compute_offset_of_blur (win).forEach (
            (offset, i) => (constraints[i].offset = offset)
        )
    }

    update_blur_effect (actor) {
        var _a
        const win = actor.meta_window
        const _effect =
            (_a = this.blur_actors.get (win)) === null || _a === void 0
                ? void 0
                : _a.get_effects ()[0]
        const effect = _effect
        if (effect) {
            effect.radius =
                settings ().global_rounded_corner_settings.border_radius
            effect.sigma = settings ().blur_sigma
            actor.opacity = settings ().blurred_window_opacity
        }
    }

    _settings_changed (k) {
        const key = k
        switch (key) {
        case 'black-list':
        case 'blur-list':
        case 'skip-libadwaita-app':
        case 'skip-libhandy-app':
            for (const actor of global.get_window_actors ()) {
                const win = actor.meta_window
                const blurred = this.blur_actors.get (win) !== undefined
                const should_be_blurred = this._should_enable_effect (win)

                if (blurred && !should_be_blurred) {
                    this._remove_effect (actor)
                }
                if (!blurred && should_be_blurred) {
                    this._add_effect (actor)
                }
            }
            break
        case 'blur-sigma':
        case 'blurred-window-opacity':
        case 'custom-rounded-corner-settings':
        case 'global-rounded-corner-settings':
            this.blur_actors.forEach ((_, win) => {
                const actor = win.get_compositor_private ()
                this.update_blur_effect (actor)
                this.update_coordinates (win)
            })
            break
        }
    }
}
