// imports.gi
const GLib       = imports.gi.GLib
const Gio        = imports.gi.Gio


// --------------------------------------------------------------- [end imports]

/**
 * This function is base on getSettings() in
 * https://gitlab.gnome.org/GNOME/gnome-shell/-/blob/main/js/misc/
 * extensionUtils.js#L211
 *
 * Need this function because preferences is launched in a new process.
 */
function getSettings (schema) {
    var _a, _b, _c
    const GioSSS = Gio.SettingsSchemaSource
    const schemaDir =
        (_b =
            (_a = Gio.File.new_for_uri (import.meta.url).get_parent ()) ===
                null || _a === void 0
                ? void 0
                : _a.get_parent ()) === null || _b === void 0
            ? void 0
            : _b.get_child ('schemas')
    let schemaSource
    if (
        schemaDir === null || schemaDir === void 0
            ? void 0
            : schemaDir.query_exists (null)
    ) {
        schemaSource = GioSSS.new_from_directory (
            (_c = schemaDir.get_path ()) !== null && _c !== void 0 ? _c : '',
            GioSSS.get_default (),
            false
        )
    } else {
        schemaSource = GioSSS.get_default ()
    }

    const schemaObj =
        schemaSource === null || schemaSource === void 0
            ? void 0
            : schemaSource.lookup (schema, true)
    if (!schemaObj) throw new Error (`Schema ${schema} could not be found`)

    return new Gio.Settings ({ settings_schema: schemaObj })
}

/** This object use to store key of settings and its type string */
const type_of_keys = {}

/**
 * Simple wrapper of Gio.Settings, we will use this class to store and
 * load settings for this gnome-shell extensions.
 */
class Settings {
    constructor () {
        /** GSettings, which used to store and load settings */
        this.g_settings = getSettings (
            'org.gnome.shell.extensions.rounded-window-effect'
        )
        // Define getter and setter for properties in class for keys in
        // schemas
        this.g_settings.list_keys ().forEach ((key) => {
            // Cache type string of keys first
            const default_val = this.g_settings.get_default_value (key)
            if (default_val == null) {
                log ('Err: Key of Settings undefined: ' + key)
                return
            }
            type_of_keys[key] = default_val.get_type_string ()

            // Define getter and setter for keys
            Object.defineProperty (this, key.replace (/-/g, '_'), {
                get: () => this.g_settings.get_value (key).recursiveUnpack (),
                set: (val) => {
                    const variant =
                        type_of_keys[key] == 'a{sv}'
                            ? this._pack_val (val)
                            : new GLib.Variant (type_of_keys[key], val)
                    this.g_settings.set_value (key, variant)
                },
            })
        })
    }

    /**
     * Just a simple wrapper to this.settings.bind(), use SchemasKeys
     * to help us check source_prop
     */
    bind (source_prop, target, target_prop, flags) {
        this.g_settings.bind (source_prop, target, target_prop, flags)
    }

    // ------------------------------------------------------- [private methods]

    /**
     * this method is used to pack javascript values into GLib.Variant when type
     * of key is `a{sv}`
     *
     * @param val Javascript object to convert
     * @returns A GLib.Variant with type `a{sv}`
     */
    _pack_val (val) {
        if (val instanceof Object) {
            const packed = {}
            Object.keys (val).forEach ((k) => {
                packed[k] = this._pack_val (val[k])
            })
            return new GLib.Variant ('a{sv}', packed)
        }

        // Important: Just handler float number and unsigned int number.
        // need to add handler to signed int number if we need store signed int
        // value into GSettings in GLib.Variant

        if (typeof val == 'number') {
            if (Math.abs (val - Math.floor (val)) < 10e-20) {
                return GLib.Variant.new_uint32 (val)
            } else {
                return GLib.Variant.new_double (val)
            }
        }

        if (typeof val == 'boolean') {
            return GLib.Variant.new_boolean (val)
        }

        if (typeof val == 'string') {
            return GLib.Variant.new_string (val)
        }

        if (val instanceof Array) {
            return new GLib.Variant (
                'av',
                val.map ((i) => this._pack_val (i))
            )
        }

        throw Error ('Unknown val to packed' + val)
    }
}

/** A singleton instance of Settings */
let _settings = null

/** Access _settings by this method */
const settings = () => {
    if (_settings != null) {
        return _settings
    }
    _settings = new Settings ()
    return _settings
}

export default settings
